/*
    SPDX-FileCopyrightText: 2010 Andriy Rysin <rysin@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "xkb_helper.h"
#include "debug.h"

#include <QDebug>
#include <QDir>
#include <QElapsedTimer>
#include <QFile>
#include <QProcess>
#include <QStandardPaths>
#include <QString>
#include <QTime>
#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
#include <QX11Info>
#else
#include <QtGui/private/qtx11extras_p.h>
#endif

#include "keyboard_config.h"
#include "keyboardsettings.h"

static const char SETXKBMAP_EXEC[] = "setxkbmap";
static const char XMODMAP_EXEC[] = "xmodmap";

static bool setxkbmapNotFound = false;
static QString setxkbmapExe;

static bool xmodmapNotFound = false;
static QString xmodmapExe;

static const QString COMMAND_OPTIONS_SEPARATOR(QStringLiteral(","));

static const auto ETC_X11_XINIT_XKBMAP = QStringLiteral("/etc/X11/xinit/Xkbmap");

static QString getSetxkbmapExe()
{
    if (setxkbmapNotFound)
        return QString();

    if (setxkbmapExe.isEmpty()) {
        setxkbmapExe = QStandardPaths::findExecutable(SETXKBMAP_EXEC);
        if (setxkbmapExe.isEmpty()) {
            setxkbmapNotFound = true;
            qCCritical(KCM_KEYBOARD) << "Can't find" << SETXKBMAP_EXEC << "- keyboard layouts won't be configured";
            return QString();
        }
    }
    return setxkbmapExe;
}

static void executeXmodmap(const QString &configFileName)
{
    if (xmodmapNotFound)
        return;

    if (QFile(configFileName).exists()) {
        if (xmodmapExe.isEmpty()) {
            xmodmapExe = QStandardPaths::findExecutable(XMODMAP_EXEC);
            if (xmodmapExe.isEmpty()) {
                xmodmapNotFound = true;
                qCCritical(KCM_KEYBOARD) << "Can't find" << XMODMAP_EXEC << "- xmodmap files won't be run";
                return;
            }
        }

        qCDebug(KCM_KEYBOARD) << "Executing" << xmodmapExe << configFileName;
        const int res = QProcess::execute(xmodmapExe, QStringList{configFileName});
        if (res != 0) {
            qCCritical(KCM_KEYBOARD) << "Failed with return code:" << res;
        }
    }
}

static void restoreXmodmap()
{
    // TODO: is just home .Xmodmap enough or should system be involved too?
    //    QString configFileName = QDir("/etc/X11/xinit").filePath(".Xmodmap");
    //    executeXmodmap(configFileName);
    QString configFileName = QDir::home().filePath(QStringLiteral(".Xmodmap"));
    executeXmodmap(configFileName);
}

// TODO: make private
bool XkbHelper::runConfigLayoutCommand(const QStringList &setxkbmapCommandArguments)
{
    QElapsedTimer timer;
    timer.start();

    const auto setxkbmapExe = getSetxkbmapExe();
    if (setxkbmapExe.isEmpty()) {
        return false;
    }

    qCDebug(KCM_KEYBOARD) << "Running" << setxkbmapExe << setxkbmapCommandArguments.join(QLatin1Char(' '));

    const int res = QProcess::execute(setxkbmapExe, setxkbmapCommandArguments);
    if (res != 0) {
        qCCritical(KCM_KEYBOARD) << "Failed with return code:" << res;
        return false;
    }

    // restore Xmodmap mapping reset by setxkbmap
    qCDebug(KCM_KEYBOARD) << "Executed successfully in " << timer.elapsed() << "ms";
    restoreXmodmap();
    qCDebug(KCM_KEYBOARD) << "\t and with xmodmap" << timer.elapsed() << "ms";
    return true;
}

class XkbMapArguments {
public:
    XkbMapArguments() = default;

    XkbMapArguments(const QString &textArgs)
    {
        static const auto whitespace = QRegExp(QStringLiteral("\\s+"));
        args = textArgs.split(whitespace, Qt::SkipEmptyParts);
    }

    static XkbMapArguments fromConfig(const QString& configPath)
    {
        QFile file{configPath};
        if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
            return {};

        return {QString::fromLatin1(file.readAll())};
    }

    QStringList get(const QString &optionName) const
    {
        auto argIdx = args.indexOf(optionName);
        if (argIdx == -1 || argIdx == args.size() - 1)
            return {};

        return args.at(argIdx + 1).split(COMMAND_OPTIONS_SEPARATOR, Qt::SkipEmptyParts);
    }

    void appendSomeTo(QStringList &destination, const QString &optionName, const QStringList &additionalOptions, bool joinWithComma) const
    {
        QStringList options = get(optionName);
        if( options.size() < 1 ) {
            return;
        }
        for (auto& additionalOpt : additionalOptions)
            if (!options.contains(additionalOpt))
                options.append(additionalOpt);

        if (joinWithComma) {
            destination.append(optionName);
            destination.append(options.join(COMMAND_OPTIONS_SEPARATOR));
        }
        else {
            for (auto& option : options) {
                destination.append(optionName);
                destination.append(option);
            }
        }
    }

    void appendOptionsTo(QStringList &destination, const QStringList &additionalOptions) const
    {
        appendSomeTo(destination, QStringLiteral("-option"), additionalOptions, false);
    }

    void appendLayoutsTo(QStringList &destination, const QStringList &additionalLayouts) const
    {
        appendSomeTo(destination, QStringLiteral("-layout"), additionalLayouts, true);
    }

private:
    QStringList args;
};

bool XkbHelper::initializeKeyboardLayouts(const QList<LayoutUnit> &layoutUnits)
{
    QStringList layouts;
    QStringList variants;
    for (const auto &layoutUnit : layoutUnits) {
        layouts.append(layoutUnit.layout());
        variants.append(layoutUnit.variant());
    }

    QStringList setxkbmapCommandArguments;
    setxkbmapCommandArguments.append(QStringLiteral("-layout"));
    setxkbmapCommandArguments.append(layouts.join(COMMAND_OPTIONS_SEPARATOR));
    if (!variants.join(QString()).isEmpty()) {
        setxkbmapCommandArguments.append(QStringLiteral("-variant"));
        setxkbmapCommandArguments.append(variants.join(COMMAND_OPTIONS_SEPARATOR));
    }

    return runConfigLayoutCommand(setxkbmapCommandArguments);
}

bool XkbHelper::initializeKeyboardLayouts(KeyboardConfig &config)
{
    QStringList setxkbmapCommandArguments;
    bool xinitArgsResetRequired = !(config.configureLayouts() && config.resetOldXkbOptions());
    auto xkbMapArgs = xinitArgsResetRequired ? XkbMapArguments::fromConfig(ETC_X11_XINIT_XKBMAP) : XkbMapArguments();
    if (!config.keyboardModel().isEmpty()) {
        XkbConfig xkbConfig;
        X11Helper::getGroupNames(QX11Info::display(), &xkbConfig, X11Helper::MODEL_ONLY);
        if (xkbConfig.keyboardModel != config.keyboardModel()) {
            setxkbmapCommandArguments.append(QStringLiteral("-model"));
            setxkbmapCommandArguments.append(config.keyboardModel());
        }
    }
    if (config.configureLayouts()) {
        QStringList layouts;
        QStringList variants;
        const QList<LayoutUnit> defaultLayouts = config.getDefaultLayouts();
        for (const auto &layoutUnit : defaultLayouts) {
            layouts.append(layoutUnit.layout());
            variants.append(layoutUnit.variant());
        }

        setxkbmapCommandArguments.append(QStringLiteral("-layout"));
        setxkbmapCommandArguments.append(layouts.join(COMMAND_OPTIONS_SEPARATOR));
        if (!variants.join(QString()).isEmpty()) {
            setxkbmapCommandArguments.append(QStringLiteral("-variant"));
            setxkbmapCommandArguments.append(variants.join(COMMAND_OPTIONS_SEPARATOR));
        }
    } else {
        xkbMapArgs.appendLayoutsTo(setxkbmapCommandArguments, {});
    }
    if (config.resetOldXkbOptions()) {
        // Pass -option "" to clear previously set options
        setxkbmapCommandArguments.append(QStringLiteral("-option"));
        setxkbmapCommandArguments.append(QStringLiteral(""));
    } else {
        xkbMapArgs.appendOptionsTo(setxkbmapCommandArguments, config.xkbOptions());
    }
    const QStringList xkbOpts = config.xkbOptions();
    for (const auto &option : xkbOpts) {
        setxkbmapCommandArguments.append(QStringLiteral("-option"));
        setxkbmapCommandArguments.append(option);
    }

    if (!setxkbmapCommandArguments.isEmpty()) {
        return runConfigLayoutCommand(setxkbmapCommandArguments);
        if (config.configureLayouts()) {
            X11Helper::setDefaultLayout();
        }
    }
    return false;
}
