/***************************************************************************
 *   Copyright (C) 2002 Joseph Wenninger <jowenn@kde.org>                  *
 *   Copyright (C) 2020 Méven Car <meven.car@kdemail.net>                  *
 *   Copyright (C) 2020 Tobias Fella <fella@posteo.de>                     *
 *   Copyright (C) 2021 Sergey V Turchin <zerg@altlinux.org>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

#include "componentchooser.h"

#include <QDBusConnection>
#include <QDBusMessage>
#include <QIcon>

#include <KApplicationTrader>
#include <KConfigGroup>
#include <KOpenWithDialog>
#include <KQuickAddons/ConfigModule>
#include <KService>
#include <KSharedConfig>
#include <KServiceTypeTrader>

ComponentChooser::ComponentChooser(QObject *parent, const QString &mimeType, const QString &type, const QString &defaultApplication, const QString &dialogText)
    : QObject(parent)
    , m_mimeType(mimeType)
    , m_type(type)
    , m_defaultApplication(defaultApplication)
    , m_dialogText(dialogText)
{
}

void ComponentChooser::defaults()
{
    if (m_defaultIndex) {
        select(*m_defaultIndex);
    }
}

void ComponentChooser::load()
{
    m_applications.clear();

    bool preferredServiceAdded = false;
    KService::Ptr preferredService, preferredServiceNull;

    QString type_icon;
    if( m_type == QStringLiteral("Audio") ) {
	type_icon = QStringLiteral("audio-x-generic");
	preferredService = KApplicationTrader::preferredService(QStringLiteral("audio/mpeg"));
    } else if( m_type == QStringLiteral("Video") ) {
	type_icon = QStringLiteral("video-x-generic");
	preferredService = KApplicationTrader::preferredService(QStringLiteral("video/x-matroska"));
    } else if( m_type == QStringLiteral("Viewer") ) {
	type_icon = QStringLiteral("image-x-generic");
	preferredService = KApplicationTrader::preferredService(QStringLiteral("image/jpeg"));
    }
    QVariantMap null_app;
    null_app["name"] = m_dialogText;
    null_app["icon"] = QIcon::fromTheme(type_icon);
    null_app["storageId"] = QLatin1String("");
    m_applications += null_app;
    m_index = m_applications.length() - 1;

    QMimeDatabase mimedb;
    for(QMimeType q_mime_type: mimedb.allMimeTypes())
    {
	QString mimetype(q_mime_type.name());
	if( m_type == QStringLiteral("Audio") && mimetype.startsWith(QString::fromLatin1("audio/")) ) {
	    if( preferredService ) {
		KService::Ptr tmpPreferredService = KApplicationTrader::preferredService(mimetype);
		if( tmpPreferredService ) {
		    if( preferredService->storageId() != tmpPreferredService->storageId() ) {
			preferredService = preferredServiceNull;
			break;
		    }
		}
	    }
	} else if( m_type == QStringLiteral("Video") &&  mimetype.startsWith(QString::fromLatin1("video/")) ) {
	    if( preferredService ) {
		KService::Ptr tmpPreferredService = KApplicationTrader::preferredService(mimetype);
		if( tmpPreferredService ) {
		    if( preferredService->storageId() != tmpPreferredService->storageId() ) {
			preferredService = preferredServiceNull;
			break;
		    }
		}
	    }
	} else if(m_type == QStringLiteral("Viewer") &&  mimetype.startsWith(QString::fromLatin1("image/")) ) {
	    if( preferredService ) {
		KService::Ptr tmpPreferredService = KApplicationTrader::preferredService(mimetype);
		if( tmpPreferredService ) {
		    if( preferredService->storageId() != tmpPreferredService->storageId() ) {
			preferredService = preferredServiceNull;
			break;
		    }
		}
	    }
	}
    }

    const KService::List services  = KServiceTypeTrader::self()->query(QLatin1String("Application"));
    for(KService::Ptr service_ptr: services) {
	KService *service(service_ptr.data());
	if( service->exec().isEmpty() )
	    continue;
	if( m_type == QStringLiteral("Audio") && service->categories().contains("Player") && service->categories().contains("Audio") ) {
	    bool have_audio = false;
	    for(QString mimetype: service->mimeTypes()) {
	        if( mimetype.startsWith(QString::fromLatin1("audio/")) ) {
	    	    have_audio = true;
		    break;
		}
	    }
	    if( !have_audio ) {
		continue;
	    }
	} else if( m_type == QStringLiteral("Video") && service->categories().contains("Player") && (service->categories().contains("AudioVideo") || service->categories().contains("Video")) ) {
	    bool have_video = false;
	    for(QString mimetype: service->mimeTypes()) {
	        if( mimetype.startsWith(QString::fromLatin1("video/")) ) {
		    have_video = true;
		    break;
		}
	    }
	    if( !have_video ) {
		continue;
	    }
	} else if( m_type == QStringLiteral("Viewer") && service->categories().contains("Viewer") && service->categories().contains("Graphics") ) {
	    bool have_image = false;
	    for(QString mimetype: service->mimeTypes()) {
	        if( mimetype.startsWith(QString::fromLatin1("image/")) ) {
		    have_image = true;
		    break;
		}
	    }
	    if( !have_image ) {
		continue;
	    }
	} else  {
	    continue;
	}
        QVariantMap application;
        application["name"] = QString(service->name() + QStringLiteral(" (") + service->genericName() + QStringLiteral(")"));
        application["icon"] = service->icon();
        application["storageId"] = service->storageId();
        m_applications += application;
        if ((preferredService && preferredService->storageId() == service->storageId())) {
            m_index = m_applications.length() - 1;
            preferredServiceAdded = true;
        }
        if (service->storageId() == m_defaultApplication) {
            m_defaultIndex = m_applications.length() - 1;
        }
    }
    if (preferredService && !preferredServiceAdded) {
        // standard application was specified by the user
        QVariantMap application;
        application["name"] = preferredService->name();
        application["icon"] = preferredService->icon();
        application["storageId"] = preferredService->storageId();
        m_applications += application;
        m_index = m_applications.length() - 1;
    }
    QVariantMap application;
    application["name"] = i18n("Other…");
    application["icon"] = QStringLiteral("application-x-shellscript");
    application["storageId"] = QLatin1String("");
    m_applications += application;
    if (m_index == -1) {
        m_index = 0;
    }

    m_previousApplication = m_applications[m_index].toMap()["storageId"].toString();
    Q_EMIT applicationsChanged();
    Q_EMIT indexChanged();
    Q_EMIT isDefaultsChanged();
}

void ComponentChooser::select(int index)
{
    if (m_index == index && m_applications.size() != 1) {
        return;
    }
    if (index == m_applications.length() - 1) {
        KOpenWithDialog *dialog = new KOpenWithDialog(QList<QUrl>(), m_mimeType, m_dialogText, QString());
        dialog->setSaveNewApplications(true);
        dialog->setAttribute(Qt::WA_DeleteOnClose);
        connect(dialog, &KOpenWithDialog::finished, this, [this, dialog](int result) {
            if (result == QDialog::Rejected) {
                Q_EMIT indexChanged();
                Q_EMIT isDefaultsChanged();
                return;
            }

            const KService::Ptr service = dialog->service();
            // Check if the selected application is already in the list
            for (int i = 0; i < m_applications.length(); i++) {
                if (m_applications[i].toMap()["storageId"] == service->storageId()) {
                    m_index = i;
                    Q_EMIT indexChanged();
                    Q_EMIT isDefaultsChanged();
                    return;
                }
            }
            const QString icon = !service->icon().isEmpty() ? service->icon() : QStringLiteral("application-x-shellscript");
            QVariantMap application;
            application["name"] = service->name();
            application["icon"] = icon;
            application["storageId"] = service->storageId();
            application["execLine"] = service->exec();
            m_applications.insert(m_applications.length() - 1, application);
            m_index = m_applications.length() - 2;
            Q_EMIT applicationsChanged();
            Q_EMIT indexChanged();
            Q_EMIT isDefaultsChanged();
        });
        dialog->open();
    } else {
        m_index = index;
    }
    Q_EMIT indexChanged();
    Q_EMIT isDefaultsChanged();
}

void ComponentChooser::saveMimeTypeAssociation(const QString &storageId)
{
    QStringList mimetypes_list;
    QMimeDatabase mimedb;
    // collect mimetypes
    for(QMimeType q_mime_type: mimedb.allMimeTypes())
    {
	QString mimetype(q_mime_type.name());
	if( m_type == QStringLiteral("Audio") && mimetype.startsWith(QString::fromLatin1("audio/")) ) {
	    mimetypes_list.append(mimetype);
	} else if( m_type == QStringLiteral("Video") &&  mimetype.startsWith(QString::fromLatin1("video/")) ) {
	    mimetypes_list.append(mimetype);
	} else if(m_type == QStringLiteral("Viewer") &&  mimetype.startsWith(QString::fromLatin1("image/")) ) {
	    mimetypes_list.append(mimetype);
	}
    }

    KSharedConfig::Ptr profile = KSharedConfig::openConfig(QStringLiteral("mimeapps.list"), KConfig::NoGlobals, QStandardPaths::GenericConfigLocation);
    if (profile->isConfigWritable(true)) {
        KConfigGroup defaultApp(profile, "Default Applications");
        KConfigGroup addedApps(profile, QStringLiteral("Added Associations"));

    for(QString mime: mimetypes_list) {
        defaultApp.writeXdgListEntry(mime, QStringList(storageId));
        QStringList apps = addedApps.readXdgListEntry(mime);
        apps.removeAll(storageId);
        apps.prepend(storageId); // make it the preferred app, i.e first in list
        addedApps.writeXdgListEntry(mime, apps);
    }

        profile->sync();
        QDBusMessage message = QDBusMessage::createMethodCall(QStringLiteral("org.kde.klauncher5"),
                                                              QStringLiteral("/KLauncher"),
                                                              QStringLiteral("org.kde.KLauncher"),
                                                              QStringLiteral("reparseConfiguration"));
        QDBusConnection::sessionBus().send(message);
    }
    m_previousApplication = m_applications[m_index].toMap()["storageId"].toString();
}

bool ComponentChooser::isDefaults() const
{
    return !m_defaultIndex.has_value() || *m_defaultIndex == m_index;
}

bool ComponentChooser::isSaveNeeded() const
{
    return !m_applications.isEmpty() && (m_previousApplication != m_applications[m_index].toMap()["storageId"].toString());
}
